@@ -1,6 +1,6 @@
 /*
     gendef - Generate list of exported symbols from a Portable Executable.
-    Copyright (C) 2009, 2010, 2011, 2012, 2013  mingw-w64 project
+    Copyright (C) 2009-2014  mingw-w64 project
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -57,6 +57,7 @@ static sImportname *theImports = NULL;
 static void *map_va (uint32_t va);
 static int is_data (uint32_t va);
 static int is_reloc (uint32_t va);
+static int has_atdecoration (void);
 
 static int disassembleRetIntern (uint32_t pc, uint32_t *retpop, sAddresses *seen, sAddresses *stack,
 				 int *hasret, int *atleast_one, const char *name, sImportname **ppimpname);
@@ -203,7 +204,7 @@ show_usage (void)
 int main(int argc,char **argv)
 {
   int i;
-  Gendefopts *opt;
+  Gendefopts *opt, *next;
 
   if (argc < 2)
   {
@@ -241,11 +242,13 @@ int main(int argc,char **argv)
 	  free (gDta);
 	  gDta = NULL;
 	}
+      next = opt->next;
       free(opt->fninput);
       free(opt->fnoutput);
       free(opt);
-      opt = opt->next;
+      opt = next;
     }
+  imp32_free ();
   return 0;
 }
 
@@ -430,7 +433,7 @@ static void *
 map_va (uint32_t va)
 {
   PIMAGE_SECTION_HEADER sec;
-  uint32_t sec_cnt,i;
+  uint32_t sec_cnt,sz,i;
   char *dptr;
 
   if (gPEDta)
@@ -445,7 +448,9 @@ map_va (uint32_t va)
     }
   for (i = 0;i < sec_cnt;i++)
     {
-      if (va >= sec[i].VirtualAddress && va < (sec[i].VirtualAddress+sec[i].Misc.VirtualSize))
+      sz = sec[i].Misc.VirtualSize;
+      if (!sz) sz = sec[i].SizeOfRawData;
+      if (va >= sec[i].VirtualAddress && va < (sec[i].VirtualAddress+sz))
         {
           dptr = (char *) &gDta[va-sec[i].VirtualAddress+sec[i].PointerToRawData];
           return (void *)dptr;
@@ -691,7 +696,7 @@ dump_def (void)
       if (exp->be64)
         exp->retpop = 0;
 
-      if (exp->retpop != (uint32_t) -1 && !exp->be64)
+      if (exp->retpop != (uint32_t) -1 && !exp->be64 && has_atdecoration())
         {
           if (exp->name[0]=='?')
             fprintf(fp," ; has WINAPI (@%u)", (unsigned int) exp->retpop);
@@ -705,7 +710,7 @@ dump_def (void)
       if (exp->beData)
         fprintf(fp," DATA");
 
-      if (exp->retpop != (uint32_t) -1 || (exp->be64 && exp->retpop == 0))
+      if (exp->retpop != (uint32_t) -1 || (exp->retpop == 0 && exp->be64) || !has_atdecoration ())
 	{
 	}
       else if (pimpname)
@@ -780,6 +785,12 @@ pop_addr (sAddresses *ad, uint32_t *val)
   return 1;
 }
 
+static int
+has_atdecoration (void)
+{
+  return (gPEDta && gPEDta->FileHeader.Machine == 0x014c /* IMAGE_FILE_MACHINE_I386 */);
+}
+
 /* exp->beData */
 static int
 disassembleRet (uint32_t func, uint32_t *retpop, const char *name, sImportname **ppimpname, int *seen_ret)
@@ -790,6 +801,11 @@ disassembleRet (uint32_t func, uint32_t *retpop, const char *name, sImportname *
   int hasret = 0;
   int atleast_one = 0;
 
+  if (!has_atdecoration())
+  {
+    *retpop = 0;
+    return 0;
+  }
   *retpop = (uint32_t) -1;
   push_addr (stack, func);
 
